@@ -3,7 +3,7 @@
  * ts_cache.h
  *	  Tsearch related object caches.
  *
- * Portions Copyright (c) 1996-2021, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2023, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * src/include/tsearch/ts_cache.h
@@ -13,7 +13,7 @@
 #ifndef TS_CACHE_H
 #define TS_CACHE_H
 
-#include "utils/guc.h"
+#include "fmgr.h"
 
 
 /*
@@ -84,7 +84,7 @@ typedef struct
 /*
  * GUC variable for current configuration
  */
-extern char *TSCurrentConfig;
+extern PGDLLIMPORT char *TSCurrentConfig;
 
 
 extern TSParserCacheEntry *lookup_ts_parser_cache(Oid prsId);
@@ -92,7 +92,5 @@ extern TSDictionaryCacheEntry *lookup_ts_dictionary_cache(Oid dictId);
 extern TSConfigCacheEntry *lookup_ts_config_cache(Oid cfgId);
 
 extern Oid	getTSCurrentConfig(bool emitError);
-extern bool check_TSCurrentConfig(char **newval, void **extra, GucSource source);
-extern void assign_TSCurrentConfig(const char *newval, void *extra);
 
 #endif							/* TS_CACHE_H */
