@@ -6,7 +6,7 @@
  *
  * This code is released under the terms of the PostgreSQL License.
  *
- * Portions Copyright (c) 1996-2022, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2023, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * src/test/regress/regress.c
@@ -39,6 +39,7 @@
 #include "parser/parse_coerce.h"
 #include "port/atomics.h"
 #include "storage/spin.h"
+#include "utils/array.h"
 #include "utils/builtins.h"
 #include "utils/geo_decls.h"
 #include "utils/lsyscache.h"
@@ -56,22 +57,22 @@
 
 #define EXPECT_EQ_U32(result_expr, expected_expr)	\
 	do { \
-		uint32		result = (result_expr); \
-		uint32		expected = (expected_expr); \
-		if (result != expected) \
+		uint32		actual_result = (result_expr); \
+		uint32		expected_result = (expected_expr); \
+		if (actual_result != expected_result) \
 			elog(ERROR, \
 				 "%s yielded %u, expected %s in file \"%s\" line %u", \
-				 #result_expr, result, #expected_expr, __FILE__, __LINE__); \
+				 #result_expr, actual_result, #expected_expr, __FILE__, __LINE__); \
 	} while (0)
 
 #define EXPECT_EQ_U64(result_expr, expected_expr)	\
 	do { \
-		uint64		result = (result_expr); \
-		uint64		expected = (expected_expr); \
-		if (result != expected) \
+		uint64		actual_result = (result_expr); \
+		uint64		expected_result = (expected_expr); \
+		if (actual_result != expected_result) \
 			elog(ERROR, \
 				 "%s yielded " UINT64_FORMAT ", expected %s in file \"%s\" line %u", \
-				 #result_expr, result, #expected_expr, __FILE__, __LINE__); \
+				 #result_expr, actual_result, #expected_expr, __FILE__, __LINE__); \
 	} while (0)
 
 #define LDELIM			'('
@@ -182,6 +183,11 @@ widget_in(PG_FUNCTION_ARGS)
 			coord[i++] = p + 1;
 	}
 
+	/*
+	 * Note: DON'T convert this error to "soft" style (errsave/ereturn).  We
+	 * want this data type to stay permanently in the hard-error world so that
+	 * it can be used for testing that such cases still work reasonably.
+	 */
 	if (i < NARGS)
 		ereport(ERROR,
 				(errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
@@ -644,9 +650,7 @@ PG_FUNCTION_INFO_V1(get_environ);
 Datum
 get_environ(PG_FUNCTION_ARGS)
 {
-#if !defined(WIN32) || defined(_MSC_VER)
 	extern char **environ;
-#endif
 	int			nvals = 0;
 	ArrayType  *result;
 	Datum	   *env;
@@ -659,7 +663,7 @@ get_environ(PG_FUNCTION_ARGS)
 	for (int i = 0; i < nvals; i++)
 		env[i] = CStringGetTextDatum(environ[i]);
 
-	result = construct_array(env, nvals, TEXTOID, -1, false, TYPALIGN_INT);
+	result = construct_array_builtin(env, nvals, TEXTOID);
 
 	PG_RETURN_POINTER(result);
 }
@@ -1101,56 +1105,6 @@ test_opclass_options_func(PG_FUNCTION_ARGS)
 	PG_RETURN_NULL();
 }
 
-/* one-time tests for encoding infrastructure */
-PG_FUNCTION_INFO_V1(test_enc_setup);
-Datum
-test_enc_setup(PG_FUNCTION_ARGS)
-{
-	/* Test pg_encoding_set_invalid() */
-	for (int i = 0; i < _PG_LAST_ENCODING_; i++)
-	{
-		char		buf[2],
-					bigbuf[16];
-		int			len,
-					mblen,
-					valid;
-
-		if (pg_encoding_max_length(i) == 1)
-			continue;
-		pg_encoding_set_invalid(i, buf);
-		len = strnlen(buf, 2);
-		if (len != 2)
-			elog(WARNING,
-				 "official invalid string for encoding \"%s\" has length %d",
-				 pg_enc2name_tbl[i].name, len);
-		mblen = pg_encoding_mblen(i, buf);
-		if (mblen != 2)
-			elog(WARNING,
-				 "official invalid string for encoding \"%s\" has mblen %d",
-				 pg_enc2name_tbl[i].name, mblen);
-		valid = pg_encoding_verifymbstr(i, buf, len);
-		if (valid != 0)
-			elog(WARNING,
-				 "official invalid string for encoding \"%s\" has valid prefix of length %d",
-				 pg_enc2name_tbl[i].name, valid);
-		valid = pg_encoding_verifymbstr(i, buf, 1);
-		if (valid != 0)
-			elog(WARNING,
-				 "first byte of official invalid string for encoding \"%s\" has valid prefix of length %d",
-				 pg_enc2name_tbl[i].name, valid);
-		memset(bigbuf, ' ', sizeof(bigbuf));
-		bigbuf[0] = buf[0];
-		bigbuf[1] = buf[1];
-		valid = pg_encoding_verifymbstr(i, bigbuf, sizeof(bigbuf));
-		if (valid != 0)
-			elog(WARNING,
-				 "trailing data changed official invalid string for encoding \"%s\" to have valid prefix of length %d",
-				 pg_enc2name_tbl[i].name, valid);
-	}
-
-	PG_RETURN_VOID();
-}
-
 /*
  * Call an encoding conversion or verification function.
  *
@@ -1185,7 +1139,7 @@ test_enc_conversion(PG_FUNCTION_ARGS)
 	int			convertedbytes;
 	int			dstlen;
 	Datum		values[2];
-	bool		nulls[2];
+	bool		nulls[2] = {0};
 	HeapTuple	tuple;
 
 	if (src_encoding < 0)
@@ -1274,7 +1228,6 @@ test_enc_conversion(PG_FUNCTION_ARGS)
 		pfree(dst);
 	}
 
-	MemSet(nulls, 0, sizeof(nulls));
 	values[0] = Int32GetDatum(convertedbytes);
 	values[1] = PointerGetDatum(retval);
 	tuple = heap_form_tuple(tupdesc, values, nulls);
@@ -1282,145 +1235,6 @@ test_enc_conversion(PG_FUNCTION_ARGS)
 	PG_RETURN_DATUM(HeapTupleGetDatum(tuple));
 }
 
-/* Convert bytea to text without validation for corruption tests from SQL. */
-PG_FUNCTION_INFO_V1(test_bytea_to_text);
-Datum
-test_bytea_to_text(PG_FUNCTION_ARGS)
-{
-	PG_RETURN_TEXT_P(PG_GETARG_BYTEA_PP(0));
-}
-
-/* And the reverse. */
-PG_FUNCTION_INFO_V1(test_text_to_bytea);
-Datum
-test_text_to_bytea(PG_FUNCTION_ARGS)
-{
-	PG_RETURN_BYTEA_P(PG_GETARG_TEXT_PP(0));
-}
-
-/* Corruption tests in C. */
-PG_FUNCTION_INFO_V1(test_mblen_func);
-Datum
-test_mblen_func(PG_FUNCTION_ARGS)
-{
-	const char *func = text_to_cstring(PG_GETARG_BYTEA_PP(0));
-	const char *encoding = text_to_cstring(PG_GETARG_BYTEA_PP(1));
-	text	   *string = PG_GETARG_BYTEA_PP(2);
-	int			offset = PG_GETARG_INT32(3);
-	const char *data = VARDATA_ANY(string);
-	size_t		size = VARSIZE_ANY_EXHDR(string);
-	int			result = 0;
-
-	if (strcmp(func, "pg_mblen_unbounded") == 0)
-		result = pg_mblen_unbounded(data + offset);
-	else if (strcmp(func, "pg_mblen_cstr") == 0)
-		result = pg_mblen_cstr(data + offset);
-	else if (strcmp(func, "pg_mblen_with_len") == 0)
-		result = pg_mblen_with_len(data + offset, size - offset);
-	else if (strcmp(func, "pg_mblen_range") == 0)
-		result = pg_mblen_range(data + offset, data + size);
-	else if (strcmp(func, "pg_encoding_mblen") == 0)
-		result = pg_encoding_mblen(pg_char_to_encoding(encoding), data + offset);
-	else
-		elog(ERROR, "unknown function");
-
-	PG_RETURN_INT32(result);
-}
-
-PG_FUNCTION_INFO_V1(test_text_to_wchars);
-Datum
-test_text_to_wchars(PG_FUNCTION_ARGS)
-{
-	const char *encoding_name = text_to_cstring(PG_GETARG_BYTEA_PP(0));
-	text	   *string = PG_GETARG_TEXT_PP(1);
-	const char *data = VARDATA_ANY(string);
-	size_t		size = VARSIZE_ANY_EXHDR(string);
-	pg_wchar   *wchars = palloc(sizeof(pg_wchar) * (size + 1));
-	Datum	   *datums;
-	int			wlen;
-	int			encoding;
-
-	encoding = pg_char_to_encoding(encoding_name);
-	if (encoding < 0)
-		elog(ERROR, "unknown encoding name: %s", encoding_name);
-
-	if (size > 0)
-	{
-		datums = palloc(sizeof(Datum) * size);
-		wlen = pg_encoding_mb2wchar_with_len(encoding,
-											 data,
-											 wchars,
-											 size);
-		Assert(wlen >= 0);
-		Assert(wlen <= size);
-		Assert(wchars[wlen] == 0);
-
-		for (int i = 0; i < wlen; ++i)
-			datums[i] = UInt32GetDatum(wchars[i]);
-	}
-	else
-	{
-		datums = NULL;
-		wlen = 0;
-	}
-
-	PG_RETURN_ARRAYTYPE_P(construct_array_builtin(datums, wlen, INT4OID));
-}
-
-PG_FUNCTION_INFO_V1(test_wchars_to_text);
-Datum
-test_wchars_to_text(PG_FUNCTION_ARGS)
-{
-	const char *encoding_name = text_to_cstring(PG_GETARG_BYTEA_PP(0));
-	ArrayType  *array = PG_GETARG_ARRAYTYPE_P(1);
-	Datum	   *datums;
-	bool	   *nulls;
-	char	   *mb;
-	text	   *result;
-	int			wlen;
-	int			bytes;
-	int			encoding;
-
-	encoding = pg_char_to_encoding(encoding_name);
-	if (encoding < 0)
-		elog(ERROR, "unknown encoding name: %s", encoding_name);
-
-	deconstruct_array_builtin(array, INT4OID, &datums, &nulls, &wlen);
-
-	if (wlen > 0)
-	{
-		pg_wchar   *wchars = palloc(sizeof(pg_wchar) * wlen);
-
-		for (int i = 0; i < wlen; ++i)
-		{
-			if (nulls[i])
-				elog(ERROR, "unexpected NULL in array");
-			wchars[i] = DatumGetInt32(datums[i]);
-		}
-
-		mb = palloc(pg_encoding_max_length(encoding) * wlen + 1);
-		bytes = pg_encoding_wchar2mb_with_len(encoding, wchars, mb, wlen);
-	}
-	else
-	{
-		mb = "";
-		bytes = 0;
-	}
-
-	result = palloc(bytes + VARHDRSZ);
-	SET_VARSIZE(result, bytes + VARHDRSZ);
-	memcpy(VARDATA(result), mb, bytes);
-
-	PG_RETURN_TEXT_P(result);
-}
-
-PG_FUNCTION_INFO_V1(test_valid_server_encoding);
-Datum
-test_valid_server_encoding(PG_FUNCTION_ARGS)
-{
-	PG_RETURN_BOOL(pg_valid_server_encoding(text_to_cstring(PG_GETARG_TEXT_PP(0))) >= 0);
-}
-
 /* Provide SQL access to IsBinaryCoercible() */
 PG_FUNCTION_INFO_V1(binary_coercible);
 Datum
